/* com.c */

/* Modul pro síťovou komunikaci. */

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <pthread.h>
#include <unistd.h>

#include "kernel.h"


/* 
 * listen on given port 
 */
int kernel_listen(uint16_t port) {
  int sock;
  int opt;
  struct sockaddr_in addr_local;

  /* create connection socket */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    printf("Chyba pri vytvareni socketu: %s\n", strerror(errno));
    return -1;
  }

  /* reusing socket */
  opt = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1) {
    printf("Varovani - nelze nastavit SO_REUSEADDR: %s\n", strerror(errno));
  }

  /* bind with local address and port */
  memset(&addr_local, 0, sizeof(addr_local));
  addr_local.sin_family = AF_INET;
  addr_local.sin_port = htons(port);
  addr_local.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock, (struct sockaddr *)&addr_local, sizeof(addr_local)) == -1) {
    printf("Chyba pri pouziti lokalni adresy: %s\n", strerror(errno));
    goto exception0;
  }

  /* prepare connection */
  if (listen(sock, 1) < 0) {
    printf("Nelze pripravit spojeni : %s\n", strerror(errno));
    goto exception0;
  }

  /* result */
  return sock;

exception0:
  close(sock);
  return -1;
}


/*poslani zpravy*/
int kernel_send(int soket, message_t *message) {
  int size;

  if((size = send(soket, message, message->size + SIZEOF_MSG_HEADER, 0)) == -1) {
    fprintf(stdout, "Chyba pri odesilani (kernel_send) %d: %s\n", errno, strerror(errno));
    return(-1);
  }
  if (size < 0) {
    fprintf(stdout, "Chyba pri odesilani (kernel_send) %d: %s\n", errno, strerror(errno));
    return(-2);
  }
  if((unsigned int)size != (message->size + SIZEOF_MSG_HEADER)) {
    fprintf(stdout, "Chyba pri odesilani (kernel_send): pouze %d bytů z %d\n", size, message->size + SIZEOF_MSG_HEADER);
    return(-3);
  }
  return(1);
}


/* 
 * odeslani retezce
 */
int kernel_strsend(int sock, const char *msg) {
  int len, size;

  if (!msg || !msg[0]) return 1;
  len = strlen(msg)/* + 1*/;

  if((size = send(sock, msg, len, 0)) < 0) {
    fprintf(stdout, "Chyba pri odesilani (kernel_strsend) %d: %s\n", errno, strerror(errno));
    return 0;
  }
  if(size != len) {
    fprintf(stdout, "Chyba pri odesilani (kernel_strsend): pouze %d bytů z %d\n", size, len);
    return 0;
  }
  
  return 1;
}


/*
 * pomocna funkce pro prijem zpravy s kontrolou
 */
static int com_recv(const char *err, int sock, char *buffer, int len) {
  int size;
 
  if ((size = recv(sock, buffer, len, 0)) < 0) {
    printf("%s: %s\n", err, strerror(errno));
    return -1;
  }
  /* nulova velikost bude chyba */
  if (!size) {
    printf("%s: Prijato 0 bytu.\n", err);
    return -1;
  }
  if (size < len) {
    printf("%s: Prilis malo dat! Minimum: %u B, prijato %u B\n", err, len, size);
    return 0;
  }

  return 1;
}


/*
 * cekani na data
 */
int kernel_select(int soket, int millis, int round) {
  struct timeval tv;
  fd_set myset;
  int i, ret = -1;

  for(i = 0; round <= 0 || i < round; i++) {
    tv.tv_sec = (int) millis/1000;   /*pocet sekund*/
    tv.tv_usec = (millis%1000)*1000; /*pocet mikrosekund*/

    FD_ZERO(&myset);
    FD_SET(soket, &myset);
    ret = select(soket + 1, &myset, NULL, NULL, &tv);
    if (ret==1) break;
    if (ret < 0) {
      printf("kernel_select: Chyba prijmu: %s\n", strerror(errno));
      return -1;
    }
  }

  return ret;
}


/*
 * prijem zpravy
 *
 * pro chybe vraci NULL (napr. ztrata spojeni),
 * pri vyprseni timeoutu vraci prazdnou zpravu
 */
message_t *kernel_recv(int soket, int millis, int round) {
  char buf[3];
  message_t *msg;
  uint16_t datasize;

  /* cekani */
  switch (kernel_select(soket, millis, round)) {
    case 0: goto exception_zero;
    case 1: break;
    default: goto exception_error;
  }

  /* hlavicka zpravy */
  switch (com_recv("Chyba prijmu hlavicky zpravy", soket, buf, sizeof(buf))) {
    case 0: goto exception_zero;
    case 1: break;
    default: goto exception_error;
  }
  memcpy(&datasize, &buf[sizeof(uint8_t)], sizeof(uint16_t));

  /* vytvoreni zpravy */  
  msg = malloc(datasize + SIZEOF_MSG_HEADER);
  memcpy(&msg->op, &buf[0], sizeof(msg->op));
  msg->size = datasize;

  /* cteni dat zpravy */
  if (datasize)
    switch (com_recv("Chyba prijmu dat zpravy", soket, &msg->data, datasize)) {
      case 0:
        free(msg);	
        goto exception_zero;
      case 1: break;
      default: 
        free(msg);
        goto exception_error;
    }
  
  return(msg);

exception_error:
//  usleep(500 * millis);
  return NULL;
exception_zero:
  msg = malloc(SIZEOF_MSG_HEADER);
  msg->op = MSG_NONE;
  msg->size = 0;
  return msg;
}


/*
 * prijem textove zpravy monitoru
 *
 * pro chybe vraci NULL (napr. ztrata spojeni),
 * pri vyprseni timeoutu vraci prazdnou zpravu
 */
char *kernel_strrecv(int soket, int millis, int round) {
  char msg[512];
  char c;
  int len;
  
  /* cekani */
  switch (kernel_select(soket, millis, round)) {
    case 0: goto exception_zero;
    case 1: break;
    default: goto exception_error;
  }
//  printf("----kernel_strrecv: selected\n");

  c = '\0';
  /* prijem znaku */
  switch (len = read(soket, msg, sizeof(msg))) {
    case 0: goto exception_zero;
    case -1: goto exception_error;
    default: break;
  }
  msg[len] = '\0';

  return strdup(msg);

exception_error:
//  usleep(500 * millis);
  return NULL;
exception_zero:
  return strdup("");
}


/*
 * send simple command
 */
int kernel_send_simple_message(int sock, uint8_t op) {
  message_t *msg;

  msg = message_create(op, "");
  if (kernel_send(sock, msg) < 0) return 0;
  message_free(msg);

  return 1;
}


/*
 * send simple message and wait for header
 */
int kernel_sendreceive_header(int sock, const char *modul, uint8_t op, uint16_t addr, uint16_t *number) {
  message_t *msg;
  
  msg = message_create(op, "W", addr);
#if TRACE_MSG
  message_dump("kernel: odesila se, ", msg);
#endif
  if (kernel_send(sock, msg) < 0) goto exception_err;
  message_free(msg);
  
  if ((msg = kernel_recv(sock, TIMEOUT, -1)) == NULL) return -1;
#if TRACE_MSG
  message_dump("kernel: prijato, ", msg);
#endif
  if (!message_test(msg, modul, MSG_DB_HEADER, "W")) goto exception;
  if (!message_parse(msg, "W", number)) goto exception;
  message_free(msg);
  
  return 1;

exception_err:
  message_free(msg);
  return -1;
exception:
  message_free(msg);
  return 0;
}
