/* kernel.c */

/* Funkce jádra - inicializace, komunikace, zámky, ... */


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <pthread.h>
#ifdef _POSIX_PRIORITY_SCHELUDING
#  include <sched.h>
#endif
#include <time.h>

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "kernel.h"


/* ochrana dat */
pthread_mutex_t zamek;         /* ochrana globálních dat */
int nreaders;                  /* pocet ctenaru */

snimac_t snimace[MAX_SENSORS]; /* tabulka snímačů */
user_info_t *uzivatele = NULL; /* tabulka uzivatelu */
log_list_t logs;               /* seznam logu */
log_list_t pending_logs;       /* seznam logu k odeslani do DB */
int nsensors = 0;              /* počet snímačů */
int nusers = 0;                /* počet uzivatelu */
int konec = 0;
int sn_invalidate = 0;         /* priznak zneplatneni dat pro snimac */

/* 
 * globalni sokety - jedno vlakno muze posilat do vic soketu,
 *
 * pokud neni vlastnik, musi pred pouzitim zamknout jadro pro cteni
 * menit soket muze pouze vlastnik (pri zneplatnovani aktualniho soketu 
 * musi byt jadro min. pro cteni)
 */
int sock_db = -1;              /* soket databaze */
int sock_sn = -1;              /* soket snimace */


/* nechráněné stavové proměnné 
 *
 * 0 ... nic
 * 1 ... vlákno běží, čekání na spojení
 * 2 ... spojeno, ...
 */
//int sens_state = 0;
//int db_state = 0;

static time_t start_time;  /* cas startu systemu */


#ifdef LOG
/* debuging */
void lprintf(const char *format, ...) {
  va_list ap;

  va_start(ap, format);
  vfprintf(stdout, format, ap);
  va_end(ap);
  fflush(stdout);
}
#endif


void llprintf(int cond, const char *format, ...) {
  va_list ap;

  if (cond) {
    va_start(ap, format);
    vfprintf(stdout, format, ap);
    va_end(ap);
    fflush(stdout);
  }
}


/*
 * inicializace jadra
 */
int kernel_init() {
  /* prepare locking, 
   *  first we must fill out the struct 
   */
  pthread_mutex_init(&zamek, NULL);
  start_time = time(NULL);
  log_init(&logs);
  log_init(&pending_logs);

  return 1;
}


/*
 * uvolni data z databaze
 */
void kernel_dbfree() {
  int i;

  for (i = 0; i < nsensors; i++) {
    if (snimace[i].users) {
      free(snimace[i].users);
      snimace[i].users = NULL;
    }
    if (snimace[i].tz) {
      free(snimace[i].tz);
      snimace[i].tz = NULL;
    }
  }
  if (uzivatele) {
    free(uzivatele);
    uzivatele = NULL;
  }
}


/*
 * end of kernel
 */
void kernel_done() {
  kernel_save_logs("kernel-logy.log");
  kernel_save("kernel-data.txt");
  kernel_dbfree();
  log_done(&logs);
  log_done(&pending_logs);

  pthread_mutex_destroy(&zamek);
}


/*
 * vyvolani Oopsu kernelu :-),
 *
 * volat pouze v pripade kriticke nebo vnitrni chyby v jadru
 */
void kernel_oops(const char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  konec = 1;
}


/*
 * dobrovolne predani rizeni jinym procesum
 */
int kernel_yield() {
#ifdef _POSIX_PRIORITY_SCHELUDING
  if (sched_yield() == -1) {
    printf("kernel, varovani: Chyba pri prepinani procesu: %s\n", strerror(errno));
  } else
#else
  usleep(50000);
#endif

  return 1;
}


/*
 * zamkne zamek pro cteni dat
 */
int kernel_lock_read() {
  int err;
  
  if ((err = pthread_mutex_lock(&zamek)) != 0) goto exception;
  lprintf("pocet ctenaru: %d --> %d\n", nreaders, nreaders + 1);
  nreaders++;
  if ((err = pthread_mutex_unlock(&zamek)) != 0) goto exception;

  return 1;

exception:
  kernel_oops("Chyba pri pouziti zamku: %s\n", strerror(errno));
  return 0;
}


/*
 * odemkne zamek pro cteni dat
 */
int kernel_unlock_read() {
  int err;
  
  if ((err = pthread_mutex_lock(&zamek)) != 0) goto exception;
  lprintf("pocet ctenaru: %d --> %d\n", nreaders, nreaders - 1);
  nreaders--;
  if ((err = pthread_mutex_unlock(&zamek)) != 0) goto exception;

  return 1;
  
exception:
  kernel_oops("Chyba pri pouziti zamku: %s\n", strerror(errno));
  return 0;
}


/*
 * zamkne zamek pro zapis
 */
int kernel_lock() {
  int err;

  for (;;) {
    if ((err = pthread_mutex_lock(&zamek)) != 0) goto exception;
    if (!nreaders) break;
    if ((err = pthread_mutex_unlock(&zamek)) != 0) goto exception;
    if (!kernel_yield()) return 0;
  }

  return 1;
  
exception:
  kernel_oops("Chyba pri pouziti zamku: %s\n", strerror(err));
  return 0;
}


/*
 * odemkne zamek pro zapis
 */
int kernel_unlock() {
  int err;
  
  if ((err = pthread_mutex_unlock(&zamek)) == 0) return 1;
  else {
    kernel_oops("Chyba pri odemylani zamku: %s\n", strerror(errno));
    return 0;
  }
}

/*
 * ceka na zamknuti zamku
 */
int kernel_wait_lock() {
  while (pthread_mutex_trylock(&zamek) == 0) {
    pthread_mutex_unlock(&zamek);
    if (!kernel_yield()) return 0;
  }

  return errno == EBUSY;
}


/* ---- sitova komunikace jadra ---- */

/*
 * cekani na spojeni
 *
 * pri chybe vraci 0 (a ceka cca. TIMEOUT ms)
 */
int kernel_accept(const char *pfx, int sock_tcp) {
  struct sockaddr_in addr;
  int addr_len;
  int sock;
  int ret;

  addr_len = sizeof(addr);
  while (!konec && (ret = kernel_select(sock_tcp, TIMEOUT, 1)) == 0) {
    lprintf("%s: cekani na spojeni\n", pfx);
  }
  if (konec || ret != 1) return -1;
  if ((sock = accept(sock_tcp, (struct sockaddr *)&addr, &addr_len)) == -1) {
    printf("%s: chyba prijmu spojeni: %s\n", pfx, strerror(errno));
    usleep(500 * TIMEOUT);
    return -1;
  }

  return sock;
}


/* ---- prace s daty ---- */

/*
 * hleda informace o uzivateli s danym id
 *
 * (vyzaduje zamek pro cteni)
 */
user_info_t *kernel_search_user(jis_t id) {
  int i;

  if (!uzivatele) {
    lprintf("kernel: uzivatele == NULL\n");
    return NULL;
  }
  for (i = 0; i < nusers; i++) {
    if (memcmp(id, uzivatele[i].id_jis, sizeof(jis_t)) == 0) return uzivatele + i;
  }

  return NULL;
}


/*
 * prihlasi uzivatele v cidle
 *
 * (vyzaduje zamek pro zapis)
 */
int kernel_login_user(int pozice, jis_t user) {
  int i;

  if (!uzivatele) {
    lprintf("kernel: uzivatele == NULL\n");
    return 0;
  }
  for (i = 0; i < snimace[pozice].nusers; i++) {
    if (memcmp(user, snimace[pozice].users[i].info->id_jis, sizeof(jis_t)) == 0) {
      snimace[pozice].users[i].info->prihlasen = 1;
      snimace[pozice].users[i].info->addr_snimace = snimace[pozice].id;
      return 1;
    }
  }

  return 0;
}


/*
 * odhlasi daneho uzivatele ze vsech snimacu
 *
 * (vyzaduje zamek pro zapis)
 */
void kernel_logout_user(jis_t id) {
  int i;
  
  if (!uzivatele) {
    lprintf("kernel: uzivatele == NULL\n");
    return;
  }
  for (i = 0; i < nusers; i++) {
    if (memcmp(id, uzivatele[i].id_jis, sizeof(jis_t)) == 0) {
      uzivatele[i].prihlasen = 0;
      uzivatele[i].addr_snimace = 0;
    }
  }
}


/*
 * doba behu systemu
 */
char *kernel_get_uptime() {
  static char buf[100];
  time_t diff;
  int hour, min, sec;
  
  diff = time(NULL) - start_time;
  sec = diff % 60, diff /= 60;
  min = diff % 60, hour = diff / 60;
  memset(buf, 0, sizeof(buf));
  snprintf(buf, sizeof(buf) - 1, "%02d:%02d:%02d", hour, min, sec);

  return buf;
}
