/* kernel.h */

#ifndef KERNEL_H_INCLUDED
#define KERNEL_H_INCLUDED

#include <inttypes.h>
#include <pthread.h>

#include "messages.h"

#define ACCESS_RIGHTS 1
#define LOCK_RIGHTS 1
#define UNLOCK_RIGHTS 1
#define DELIVER_RIGHTS 1

#define MAX_SENSORS 256
#define TIMEOUT 4000

/* logging */
#define LOG 1
#define TRACE_MSG 1


typedef struct {
  int id_user, id_group;
  jis_t id_jis;
  uint8_t key[16];
  char jmeno[16], prijmeni[16], os_cislo[16];
  int prihlasen;
  uint16_t addr_snimace;
} user_info_t;

typedef struct {
  uint8_t id_zone;
  int rights;
  user_info_t *info;
} user_t;

typedef struct {
  uint8_t id;
  uint8_t bmp[BITMAP_SIZE];
} time_zone_t;

typedef struct log_s log_t;

typedef struct log_s {
  log_t *next;
  uint16_t sensor;
  int type;           /* typ udalosti ze snimace (-1 = vlastni) */
  jis_t user;         /* uzivatel */
  char time[17];      /* case ve forme textu */
  char text[256];     /* popis udalosti */
} log_s;

typedef struct {
  log_t *logs;
  log_t *last, *current;
} log_list_t;

typedef struct {
  uint16_t id;
  user_t *users;
  time_zone_t *tz;  
  int nusers, nzones;
  int room;
  int active;
} snimac_t;

typedef struct list_item_s list_item_t;

typedef struct list_item_s {
  list_item_t *next;
  char *str;
  void *data;
} list_item_s;

typedef struct {
  list_item_t *first, *current, *last;
} list_t;

typedef int (order_f(void *a, void *b, void *user_data));

extern snimac_t snimace[MAX_SENSORS]; /* tabulka snímačů */
extern user_info_t *uzivatele;        /* tabulka uzivatelu */
extern log_list_t logs;               /* seznam logu */
extern log_list_t pending_logs;       /* seznam logu k odeslani do DB */
extern int nsensors;                  /* počet snímačů */
extern int nusers;                    /* počet uzivatelu */
extern int sock_db;
extern int sock_sn;
extern int konec;
extern int sn_invalidate;             /* priznak zneplatneni dat pro snimac */


/* ---- debug ---- */
#ifdef LOG
void lprintf(const char *format, ...);
void llprintf(int cond, const char *format, ...);
#else
#define lprintf(a, ...)
#define llprintf(a, b, ...)
#endif


/* ---- komunikace ---- */

/* 
 * listen on given port 
 */
int kernel_listen(uint16_t port);

/*poslani zpravy*/
int kernel_send(int soket, message_t *message);

/* 
 * odeslani retezce
 */
int kernel_strsend(int sock, const char *msg);

/*
 * cekani na data
 */
int kernel_select(int socket, int millis, int round);

/*
 * cekani na spojeni
 *
 * pri chybe vraci 0 (a ceka cca. TIMEOUT ms)
 */
int kernel_accept(const char *pfx, int sock_tcp);

/*
 * prijem zpravy
 *
 * pro chybe vraci NULL (napr. ztrata spojeni),
 * pri vyprseni timeoutu vraci prazdnou zpravu
 */
message_t *kernel_recv(int soket, int millis, int round);

/*
 * prijem textove zpravy monitoru
 *
 * pro chybe vraci NULL (napr. ztrata spojeni),
 * pri vyprseni timeoutu vraci prazdnou zpravu
 */
char *kernel_strrecv(int soket, int millis, int round);

/*
 * send simple command
 */
int kernel_send_simple_message(int sock, uint8_t op);

/*
 * send simple message and wait for header
 */
int kernel_sendreceive_header(int sock, const char *modul, uint8_t op, uint16_t addr, uint16_t *number);


/* ---- funkce jadra, zamky, ... ---- */

/*
 * inicializace jadra
 */
int kernel_init();

/*
 * uvolni data z databaze
 */
void kernel_dbfree();

/*
 * end of kernel
 */
void kernel_done();

/*
 * vyvolani Oopsu kernelu :-),
 *
 * volat pouze v pripade kriticke nebo vnitrni chyby v jadru
 */
void kernel_oops(const char *fmt, ...);

/*
 * dobrovolne predani rizeni jinym procesum
 */
int kernel_yield();

/*
 * zamkne zamek pro cteni dat
 */
int kernel_lock_read();

/*
 * odemkne zamek pro cteni dat
 */
int kernel_unlock_read();

/*
 * zamkne zamek pro zapis
 */
int kernel_lock();

/*
 * odemkne zamek pro zapis
 */
int kernel_unlock();

/*
 * ceka na zamknuti zamku
 */
int kernel_wait_lock();


/* ---- prace s logy ---- */

/*
 * inicializace (vynulovani) seznamu logu
 */
void log_init(log_list_t *list);

/*
 * zrusi seznam zprav
 */
void log_done(log_list_t *list);

/*
 * pridani logu do seznamu
 */
log_t *log_add(log_list_t *list, uint16_t sensor, int type, uint8_t user[], const char *time, const char *text);

/*
 * vrati prvni zaznam
 */
log_t *log_first(log_list_t *list);

/*
 * vraci nasledujici zpravy v seznamu
 */
log_t *log_next(log_list_t *list);

/*
 * vraci, zda je seznam logu prazdny
 */
int log_isempty(log_list_t *list);


/* ---- snimace ---- */

int sn_recv(int sock_sn);


/* ---- databaze ---- */

/*
 * receive all info from database
 */
int db_recv(int sock);

/* 
 * komunikace s monitorem
 */
int mn_recv(int sock);


/* ---- preace se seznamem ---- */

/* inicializace seznamu */
void list_init(list_t *list);

/* zruseni seznamu */
void list_done(list_t *list);

/* vrati prvni polozku seznamu */
list_item_t *list_first(list_t *list);

/* vrati dalsi polozku seznamu */
list_item_t *list_next(list_t *list);

/* 
 * vlozeni polozky na konec seznamu 
 */
void list_add(list_t *list, char *str, void *data);

/* 
 * vlozeni polozky,
 *
 * omezeni: 'str' a 'data' budou uvolnovany pomoci free()
 */
void list_insert(list_t *list, char *str, void *data, order_f *order, void *f_data);

/* 
 * vlozeni polozky podle int,
 *
 * omezeni: 'str' budou uvolnovany pomoci free()
 */
void list_insert_by_int(list_t *list, char *str, int idata);

/* 
 * vlozeni polozky podle retezce,
 *
 * omezeni: 'str' bude uvolnovano pomoci free()
 */
void list_insert_by_str(list_t *list, char *str, char *sdata);

/* 
 * vlozeni polozky podle JIS,
 *
 * omezeni: 'str' bude uvolnovano pomoci free()
 */
void list_insert_by_jis(list_t *list, char *str, jis_t jis);


/* ---- prace s lokalnimi daty ---- */

/*
 * ulozi data jadra
 *
 * predpoklada jadro zamcene pro cteni
 */
int kernel_save(const char *name);

/* nacte data jadra ze souboru 
 *
 * predpoklada jadro zamcene pro zapis
 */
int kernel_load(const char *name);

/*
 * ulozi logy vsech snimacu,
 *
 * predpoklada jadro zamcene pri cteni
 */
int kernel_save_logs(const char *name);

/*
 * hleda informace o uzivateli s danym id
 *
 * (vyzaduje zamek pro cteni)
 */
user_info_t *kernel_search_user(jis_t id);

/*
 * prihlasi uzivatele v cidle
 *
 * (vyzaduje zamek pro zapis)
 */
int kernel_login_user(int pozice, jis_t user);

/*
 * odhlasi daneho uzivatele ze vsech snimacu
 *
 * (vyzaduje zamek pro zapis)
 */
void kernel_logout_user(jis_t id);

/*
 * doba behu systemu
 */
char *kernel_get_uptime();

#endif
