/* list.c */

/* setridene ukladani do spojoveho seznamu - pouzito pro razene vypisy */


#include <malloc.h>
#include <string.h>

#include "kernel.h"


/* inicializace seznamu */
void list_init(list_t *list) {
  list->first = NULL;
  list->current = NULL;
  list->last = NULL;
}


/* zruseni seznamu */
void list_done(list_t *list) {
  list_item_t *item;
  
  if (!list) return;
  item = list->first;
  while (item) {
    list->current = item->next;
    if (item->data) free(item->data);
    if (item->str) free(item->str);
    free(item);
    item = list->current;
  }
}


/* vrati prvni polozku seznamu */
list_item_t *list_first(list_t *list) {
  if (!list) return NULL;
  list->current = list->first;
  return list->first;
}


/* vrati dalsi polozku seznamu */
list_item_t *list_next(list_t *list) {
  if (!list || !list->current) return NULL;
  list->current = list->current->next;
  return list->current;
}


/* 
 * vlozeni polozky na konec seznamu 
 */
void list_add(list_t *list, char *str, void *data) {
  list_item_t *item;

  if (!list) return;
  item = malloc(sizeof(list_item_t));
  item->data = data;
  item->str = str;
  item->next = NULL;

  if (!list->first) {
    list->first = item;
    list->last = item;
  } else {
    list->last->next = item;
    list->last = item;
  }
}


/* 
 * vlozeni polozky,
 *
 * omezeni: 'str' a 'data' budou uvolnovany pomoci free()
 */
void list_insert(list_t *list, char *str, void *data, order_f *order, void *f_data) {
  list_item_t *item, *i1, *i2;
  
  if (!list) return;
  item = malloc(sizeof(list_item_t));
  item->data = data;
  item->str = str;

  /* seznam je prazdny nebo na zacatek seznamu */
  if (!list->first || order(data, list->first->data, f_data) < 0) {
    item->next = list->first;
    list->last = list->first;
    list->first = item;
    return;
  }

  /* vlozeni nekam za first */
  i1 = list->first;
  while ((i2 = i1->next) != NULL && order(data, i2->data, f_data) >= 0) i1 = i2;
  item->next = i2;
  i1->next = item;
  /* update last */
  if (!i2) list->last = item;
}


static int iorder(void *ag, void *bg, void *unused) {
  int a = *((int *)ag);
  int b = *((int *)bg);

  return a > b ? 1 : (a == b ? 0 : -1);
}


/* 
 * vlozeni polozky podle int,
 *
 * omezeni: 'str' budou uvolnovany pomoci free()
 */
void list_insert_by_int(list_t *list, char *str, int idata) {
  int *data;

  data = malloc(sizeof(int));
  *data = idata;
  list_insert(list, str, data, iorder, NULL);
}


static int sorder(void *ag, void *bg, void *unused) {
  return strcmp((char *)ag, (char *)bg);
}


/* 
 * vlozeni polozky podle retezce,
 *
 * omezeni: 'str' bude uvolnovano pomoci free()
 */
void list_insert_by_str(list_t *list, char *str, char *sdata) {
  list_insert(list, str, strdup(sdata), sorder, NULL);
}


static int jorder(void *ag, void *bg, void *unused) {
  return memcmp(ag, bg, sizeof(jis_t));
}


/* 
 * vlozeni polozky podle JIS,
 *
 * omezeni: 'str' bude uvolnovano pomoci free()
 */
void list_insert_by_jis(list_t *list, char *str, jis_t jdata) {
  jis_t *jis;
  
  jis = malloc(sizeof(jis_t));
  memcpy(jis, jdata, sizeof(jis_t));
  list_insert(list, str, jis, jorder, NULL);
}


#ifdef TEST

#include <stdio.h>
#include <string.h>

int main() {
  list_t list;
  list_item_t *item;

  list_init(&list);
  list_insert_by_int(&list, strdup("aaaa"), 1, order, NULL);
  list_insert_by_int(&list, strdup("cccc"), 3, order, NULL);
  list_insert_by_int(&list, strdup("bbbb"), 2, order, NULL);
  list_insert_by_int(&list, strdup("____"), 0, order, NULL);
  list_insert_by_int(&list, strdup("dddd"), 4, order, NULL);

  item = list_first(&list);
  while (item) {
    printf("%s\n", item->str);
    item = list_next(&list);
  }
  
  list_done(&list);

  return 0;
}
#endif
