/* messages.c */

/* 
 * Sestavování a zpracování zpráv.
 *
 * Endian ignorován... 
 */

#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <errno.h>

#include "kernel.h"


#define SIZE_OF_USER 35
#define SIZE_OF_TZ 49


unsigned long hex2ulong(const char *hex) {
  unsigned long ul;
  const char *str, *fmt;
  
  if (!hex || !hex[0]) return 0;
  if (hex[0] == '0' && hex[1] == 'x') {
    if (!hex[2]) {
      printf("kernel: retezec '0x' nelze prevest na cislo\n");
      return 0;
    }
    str = hex + 2;
    fmt = "%lX";
  } else {
    str = hex;
    fmt = "%lu";
  }
  
  if (sscanf(str, fmt, &ul) != 1) {
    printf("kernel: chyba pri prevodu '%s' (dle '%s'): %s\n", str, fmt, strerror(errno));
    return 0;
  }

  return ul;
}


static int format_size(const char *format) {
  int datasize;

  datasize = 0;
  if (format)
    while (*format) {
      switch(*format++) {
        case 'B': datasize++; break;
        case 'W': datasize += 2; break;
        case 'L': datasize += 4; break;
        case 'I': datasize += sizeof(jis_t); break;
        case 'K': datasize += 16; break;
        case 'T': datasize += BITMAP_SIZE; break;
        case 'S': datasize += STR_MAXLEN; break;
        default:
          printf("kernel: Chyba ve formátovacím řetězci '%s'\n", format);
          return 0;
      }
    }

  return datasize;
}


static void format_assemble(const char *format, va_list ap, char *buffer) {
  uint8_t *src, *dst;
  size_t i;
  
  while (*format) {
    switch(*format++) {
      case 'B':
        *((uint8_t *)buffer++) = va_arg(ap, int);
        break;
      case 'W':
        *((uint16_t *)buffer) = va_arg(ap, int);
        (buffer) += 2;
        break;
      case 'L':
        *((uint32_t *)buffer) = va_arg(ap, uint32_t);
        (buffer) += 4;
        break;
      case 'I':
	src = va_arg(ap, uint8_t *);
        dst = (uint8_t *)buffer;
	for (i = 0; i < sizeof(jis_t); i++) dst[i] = src[i];
	buffer += sizeof(jis_t);
	break;
      case 'K':
	src = va_arg(ap, uint8_t *);
        dst = (uint8_t *)buffer;
	for (i = 0; i < 16; i++) dst[i] = src[i];
	buffer += 16;
	break;
      case 'T':
        src = va_arg(ap, uint8_t *);
        dst = (uint8_t *)buffer;
        for (i = 0; i < BITMAP_SIZE; i++) dst[i] = src[i];
        buffer += BITMAP_SIZE;
        break;
      case 'S': {
        char *srcstr, *dststr;

        srcstr = va_arg(ap, char *);
        dststr = (char *)buffer;
        memset(dststr, 0, STR_MAXLEN);
        strncpy(dststr, srcstr, STR_MAXLEN - 1);
        buffer += STR_MAXLEN;
      }
      default:
        break;
    }
  }
}


static void format_parse(const char *format, va_list ap, char *buffer) {
  uint8_t *src, *dst;
  uint32_t *l;
  int *d;
  size_t i;
  
  while (*format) {
    switch(*format++) {
      case 'B':
        d = va_arg(ap, int *);
        if (d) *d = *((uint8_t *)buffer++);
        break;
      case 'W':
        d = va_arg(ap, int *);
        if (d) *d = *((uint16_t *)buffer);
        (buffer) += 2;
        break;
      case 'L':
        l = va_arg(ap, uint32_t *);
        if (l) *l = *((uint32_t *)buffer);
        (buffer) += 4;
        break;
      case 'I':
        src = (uint8_t *)buffer;
	dst = va_arg(ap, uint8_t *);
	for (i = 0; i < sizeof(jis_t); i++) dst[i] = src[i];
	buffer += sizeof(jis_t);
        break;
      case 'K':
        src = (uint8_t *)buffer;
	dst = va_arg(ap, uint8_t *);
	for (i = 0; i < 16; i++) dst[i] = src[i];
	buffer += 16;
        break;
      case 'T':
        src = (uint8_t *)buffer;
        dst = va_arg(ap, uint8_t *);
        for (i = 0; i < BITMAP_SIZE; i++) dst[i] = src[i];
        buffer += BITMAP_SIZE;
        break;
      case 'S': {
        char *srcstr, *dststr;

        srcstr = (char *)buffer;
        dststr = va_arg(ap, char *);
        memset(dststr, 0, STR_MAXLEN);
        strncpy(dststr, srcstr, STR_MAXLEN - 1);
        buffer += STR_MAXLEN;
      }
      default:
        break;
    }
  }
}


message_t *message_create_v(uint8_t op, const char *format, va_list ap) {
  message_t *msg;
  int datasize;

  datasize = format_size(format);

  msg = malloc(SIZEOF_MSG_HEADER + datasize);
  if (!msg) return NULL;

  msg->op = op;
  msg->size = datasize;
  if (datasize > 0)
    format_assemble(format, ap, &msg->data);

  return msg;
}


message_t *message_create_part_v(uint8_t op, uint16_t more_size, const char *format, va_list ap) {
  message_t *msg;
  uint16_t datasize;

  datasize = format_size(format);

  msg = malloc(SIZEOF_MSG_HEADER + datasize + more_size);
  if (!msg) return NULL;

  msg->op = op;
  if (datasize > 0) {  
    msg->size = datasize + more_size;
    format_assemble(format, ap, &msg->data);
  } else
    msg->size = more_size;

  return msg;
}


int message_add_part_v(message_t *msg, uint16_t pos, const char *format, va_list ap) {
  uint16_t datasize;
  
  datasize = format_size(format);
  if (pos + datasize > msg->size) {
    printf("message_add_part_v: pozice %u + data %u > velikost %u\n", pos, datasize, msg->size);
    return 0;
  }

  if (datasize)
    format_assemble(format, ap, ((char *)&msg->data) + pos);

  return 1;
}


int message_parse_v(message_t *message, const char *format, va_list ap) {
  int datasize;

  datasize = format_size(format);
  if (datasize > message->size) {
    printf("message_parse_v: zprava prilis kratka (delka %u, format '%s')\n", message->size, format);
    return 0;
  }
  if (datasize < message->size) {
    printf("message_parse_v, varovani: zprava prilis dlouha (delka %u, format '%s')\n", message->size, format);
  }

  if (datasize)
    format_parse(format, ap, &message->data);

  return 1;
}


int message_parse_part_v(message_t *message, uint16_t pos, const char *format, va_list ap) {
  int datasize;

  datasize = format_size(format);
  if (datasize > message->size) {
    printf("message_parse_part_v: zpráva příliš krátká (pos %u, len %u, formát '%s')\n", pos, message->size, format);
    return 0;
  }

  if (datasize)
    format_parse(format, ap, ((char *)&message->data) + pos);

  return 1;
}


message_t *message_create(uint8_t op, const char *format, ...) {
  va_list ap;
  message_t *msg;

  va_start(ap, format);
  msg = message_create_v(op, format, ap);
  va_end(ap);

  return msg;
}


message_t *message_create_part(uint8_t op, uint16_t more_size, const char *format, ...) {
  message_t *msg;
  va_list ap;

  va_start(ap, format);
  msg = message_create_part_v(op, more_size, format, ap);
  va_end(ap);

  return msg;
}


int message_add_part(message_t *msg, uint16_t pos, const char *format, ...) {
  va_list ap;
  int retval;
  
  va_start(ap, format);
  retval = message_add_part_v(msg, pos, format, ap);
  va_end(ap);

  return retval;
}


void message_free(message_t *message) {
  if (message) free(message);
}


int message_parse(message_t *message, const char *format, ...) {
  va_list ap;
  int retval;

  va_start(ap, format);
  retval = message_parse_v(message, format, ap);
  va_end(ap);

  return retval;
}


int message_parse_part(message_t *message, uint16_t pos, const char *format, ...) {
  va_list ap;
  int retval;

  va_start(ap, format);
  retval = message_parse_part_v(message, pos, format, ap);
  va_end(ap);

  return retval;
}


int message_test(message_t *message, const char* pfx, uint8_t op, const char *format) {
  int datasize;

  if (!message) {
    printf("%s: zpráva je NULL\n", pfx);
    return 0;
  }
  if (op != (uint8_t)-1 && message->op != op) {
    printf("%s: očekává se typ %02X, přijato %02X\n", pfx, op, message->op);
    return 0;
  }
  if (format) {
    datasize = format_size(format);
    if (datasize != message->size) {
      printf("%s: Velikost dat by měla být %u (ale je %u)", pfx, datasize, message->size);
      return 0;
    }
  }

  return 1;
}


void message_dump(const char *pref, message_t *msg) {
  int i;
  unsigned char *buf;
  
  if (msg) {
    printf("%sop=0x%02X, size=%u", pref, msg->op, msg->size);
    if (msg->size) {
      printf(", ");
      buf = &msg->data;
      for (i = 0; i < msg->size; i++) printf("%02X ", buf[i]);
    }
    printf("\n");
  } else {
    printf("%s(null)\n", pref);
  }
}


/* 
 * prevede pole bytu na hexa-string 
 */
char *id2str(uint8_t *id, size_t n) {
  static char str[2 * 16 + 3];
  char num[3];
  size_t i;

  strcpy(str, "0x00");
  if (2 * n + 3 > sizeof(str)) {
    printf("kernel: moc velke pole (%d)\n", n);
    return str;
  }
  for (i = 0; i < n; i++) {
    sprintf(num, "%02X", id[i]);
    strcpy(str + 2*i + 2, num);
  }

  return str;
}


/* 
 * prevede hexa-string na id 
 */
void str2id(const char *str, uint8_t *id, size_t n) {
  size_t i, len, u;
  char num[3];
  unsigned int c;

  len = strlen(str);
  if (len != 2 * n + 2) {
    printf("kernel: spatny format cisla: %s\n", str);
    goto exception;
  }

  i = 2;
  u = 0;
  strcpy(num, "00");
  while (i < len) {
    if (u > n) {
      printf("kernel, vnitrni chyba: cislo '%s' prilis dlouhe pro %d bytu\n", str, n);
      goto exception;
    }
    memcpy(num, str + i, 2);
    if (sscanf(num, "%X", &c) != 1) {
      printf("kernel: naka chyba pri prevodu '%s': %s\n", num, strerror(errno));
      goto exception;
    }
    if (c > 255) {
      printf("kernel, vnitrni chyba: byte %u mimo rozsah\n", c);
      goto exception;
    }
    id[u++] = c;
    i += 2;
  }

  return;

exception:
  memset(id, 0, n); 
  return;
}


#if 0
/*
 * získá 'nusers' uživatelů ze zprávy 'msg' a uloží je do 'users'
 */
static int message_get_users(message_t *msg, int nusers, user_t *users) {
  int i;
  int id_zone, rights;
  
  for (i = 0; i < nusers; i++) {
    if (!message_parse_part(msg, 2 + SIZE_OF_USER * i, "KKBW", users[i].id, users[i].key, &id_zone, &rights)) return 0;
    users[i].id_zone = id_zone;
    users[i].rights = rights;
  }
      

  return 1;
}


/*
 * získá 'ntzs' časových zón ze zprávy 'msg' a uloží je do 'tzs'
 */
static int message_get_tzs(message_t *msg, uint16_t ntzs, time_zone_t *tzs) {
  uint16_t i;
  int id;
  
  for (i = 0; i < ntzs; i++) {
    if (!message_parse_part(msg, 2 + SIZE_OF_TZ * i, "BT", &id, tzs[i].bmp))
      return 0;
    tzs[i].id = id;
  }

  return 1;
}


int message_db_users_parse(const char *modul, message_t *msg, uint16_t *index, uint16_t max_index, user_t *users) {
  uint16_t nusers_loc;
  int ee;
  
  /* test for type of frame */ 
  if (msg->op != MSG_DB_USERS) {
    printf("%s: Očekává se rámec s uživateli, typ %02X (přijato %02X)", modul, MSG_DB_USERS, msg->op);
    return 0;
  }
  /* local number of users */
  if (!message_parse_part(msg, 0, "W", &ee)) return 0;
  nusers_loc = ee;
  if (*index + nusers_loc > max_index) {
    printf("%s: Varování - uživatelů příliš mnoho (přijato %d, maximum %d)\n", modul, nusers_loc, max_index - *index);
    nusers_loc = max_index - *index;
  }
  /* parse users in the frame */
  if (!message_get_users(msg, nusers_loc, users + *index)) {
    printf("%s: Příliš málo dat pro data %d uživatelů (máme %d, nutno %d)\n", modul, nusers_loc, msg->size, sizeof(nusers_loc) + nusers_loc * SIZE_OF_USER);
    return 0;
  }
  (*index) += nusers_loc;

  return 1;
}


int message_db_tzs_parse(const char *modul, message_t *msg, uint16_t *index, uint16_t max_index, time_zone_t *tzs) {
  uint16_t ntzs_loc;
 
  /* test for type of frame */ 
  if (msg->op != MSG_DB_USERS) {
    printf("%s: Očekává se rámec s uživateli, typ %02X (přijato %02X)", modul, MSG_DB_USERS, msg->op);
    return 0;
  }
  /* local number of time zones */
  if (!message_parse_part(msg, 0, "W", &ntzs_loc)) return 0;
  if (*index + ntzs_loc > max_index) {
    printf("%s: Varování - časových zón příliš mnoho (přijato %d, maximum %d)\n", modul, ntzs_loc, max_index - *index);
    ntzs_loc = max_index - *index;
  }
  /* parse time zones in the frame */
  if (!message_get_tzs(msg, ntzs_loc, tzs + *index)) {
    printf("%s: Příliš málo dat pro %d časových zón (máme %d, nutno %d)\n", modul, ntzs_loc, msg->size, sizeof(ntzs_loc) + ntzs_loc * SIZE_OF_TZ);
    return 0;
  }
  (*index) += ntzs_loc;

  return 1;
}
#endif


#ifdef TEST
int main() {
  message_t *msg;

  {
    uint16_t a, b, d;
    uint8_t c;
    char str[STR_MAXLEN];
  
    msg = message_create(1, "WWBWS", 0x4876, 0x12C4, 0x23, 0xFEED, "ahoj");
    message_dump("test_msg1(WWBWS, 0x4876, 0x12C4, 0x23, 0xFEED, \"ahoj\"):\n  ", msg);
    message_parse(msg, "");
    message_parse(msg, NULL);
    message_parse(msg, "WWBW", NULL, NULL, NULL, NULL);
    message_parse(msg, "WWBWS", &a, &b, &c, &d, str);
    printf("  data ve zprávě: 0x%04X, 0x%04X, 0x%02X, 0x%04X, %s\n", a, b, c, d, str);
    free(msg);
    printf("\n");
  }

  {
    uint8_t a, c, d;
    uint16_t b;
    uint32_t key[] = {0x12345678, 0x87654321, 0xAABBBBAA, 0xFFEEDDBB};
    
    msg = message_create_part(1, 4 + 16, "WWB", 0x4876, 0x12C4, 0x23);
    message_add_part(msg, 5, "WBB", 0xFEEF, 0xDD, 0xCC);
    message_add_part(msg, 5 + 4, "K", key);
    message_dump("test_msg2('WWBWBB', 0x4876, 0x12C4, 0x23, 0xFEEF, 0xDD, 0xCC):\n  ", msg);
    message_parse_part(msg, 4, "");
    message_parse_part(msg, 4, NULL);
    message_parse_part(msg, 4, "BWBB", NULL, NULL, NULL, NULL);
    message_parse_part(msg, 4, "BWBB", &a, &b, &c, &d);
    printf("  data ve zprávě od poz. 4: 0x%02X, 0x%04X, 0x%02X, 0x%02X\n", a, b, c, d);
    /* test přesahu */
    message_parse_part(msg, 4, "WWBWBBKB", &a, &b, &c, &d, NULL, NULL);
    free(msg);
    printf("\n");
  }

  {
    char str[256];
    uint8_t id[16];
    int i;
    
    memset(id, 0, sizeof(id));
  
    strcpy(str, "0x00112233445566778899AABBCCDDEEFF");
    str2id(str, id, 16);
    printf("str2id(%s, id) ==> id = 0x", str);
    for (i = 0; i < 16; i++) printf("%02X", id[i]);
    printf("\n");

    memset(str, 0, sizeof(str));

    strcpy(str, id2str(id, 16));
    printf("id2str(id, 16) ==> %s\n", str);

    strcpy(str, "0xAABBCCDD");
    printf("hex2ulong(%s) ==> 0x%lX\n", str, hex2ulong(str));

    strcpy(str, "0xFF");
    printf("hex2ulong(%s) ==> 0x%lX\n", str, hex2ulong(str));
  }

  return 0;
}
#endif
