/* monitor.c */

#include <inttypes.h>

#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>

#include "kernel.h"


#define MSG_MN_USERS 0x41 /* seznam uzivatelu */
#define MSG_MN_LOGINS 0x42 /* prihlaseni uzivatele */
#define MSG_MN_SENSORS 0x43 /* seznam cidel */
#define MSG_MN_ROOM_LOGINS 0x44 /* uzivatele v mistnosti */
#define MSG_MN_LOG 0x51 /* vsechny logy */
#define MSG_MN_SERVER_STATUS 0x60 /* stav jadra */
#define MSG_MN_END 0x79 /* koncova znacka */


static char *msg_create_v(int op, const char *fmt, va_list ap) {
  char msg[1024], part[256], *str;
  unsigned int w;
  signed int i;
  uint8_t *p;

  sprintf(msg, "%d|", op);
  part[255] = '\0';
  while (*fmt) {
    switch (*fmt++) {
      case 'H':
        w = va_arg(ap, unsigned int);
	snprintf(part, sizeof(part) - 1, "0x%04X", w);
        strcat(msg, part);
	break;
      case 'w':
        w = va_arg(ap, unsigned int);
	snprintf(part, sizeof(part) - 1, "%u", w);
        strcat(msg, part);
	break;
      case 'i':
        i = va_arg(ap, signed int);
	snprintf(part, sizeof(part) - 1, "%d", i);
        strcat(msg, part);
	break;
      case 'S':
        str = va_arg(ap, char *);
	strcat(msg, str);
	break;
      case 'I':
        p = va_arg(ap, uint8_t *);
        printf("TODO\n");
        break;
      default:
        printf("mn, vnitrni chyba: formatovaci retezec %s neni platny\n", fmt);
        return NULL;
    }
    strcat(msg, "|");
  }
  strcat(msg, "\n");

  return strdup(msg);
}


static char *msg_create(int op, const char *fmt, ...) {
  va_list ap;
  char *str;

  va_start(ap, fmt);
  str = msg_create_v(op, fmt, ap);
  va_end(ap);

  return str;
}


static int msg_parse(char *msg, char *tokens[], int maxn) {
  int n;

  n = 0;
  tokens[n] = strtok(msg, "|");
  while (tokens[n] && (n + 1) < maxn) {
    n++;
    tokens[n] = strtok(NULL, "|");
  }

  return n;
}


/* odesle seznam zprav */
static int list_send(int sock, list_t *list) {
  list_item_t *item;
  char *msg;

  /* odeslani vsech pripravenych zprav */
  item = list_first(list);
  while (item) {
    llprintf(TRACE_MSG, "mn: odeslani zpravy '%s'\n", item->str);
    if (!kernel_strsend(sock, item->str)) return 0;
    item = list_next(list);
  }
  msg = msg_create(MSG_MN_END, "");
  if (!kernel_strsend(sock, msg)) {
    free(msg);
    return 0;
  }
  free(msg);

  return 1;
}


#ifndef TEST
static int parse_order(const char *order) {
  if (strcmp(order, "jmeno") == 0) return 1;
  else if (strcmp(order, "prijmeni") == 0) return 2;
  else if (strcmp(order, "os_cislo") == 0) return 3;
  else if (strcmp(order, "id_jis") == 0) return 4;
  else if (strcmp(order, "id_uzivatele") == 0) return 5;
  else if (strcmp(order, "id_skupiny") == 0) return 6;
  else {
    printf("mn, varovani: nezname trideni '%s'\n", order);
    return 0;
  }
}


static int monitor_send_loglist(int sock, log_list_t *loglist) {
  char *msg;
  log_t *log;

  log = log_first(loglist);
  while (log) {
    user_info_t *ui;
    
    ui = kernel_search_user(log->user);
    msg = msg_create(MSG_MN_LOG, "wSSS", log->type, ui ? ui->os_cislo : "-", log->time, log->text);
    if (!msg) {
      printf("mn, vnitrni chyba: neslo vytvorit zpravu u snimace %04X\n", log->sensor);
      return 0;
    }
    llprintf(TRACE_MSG, "mn: sending MSG_MN_LOG (%s)...", msg);
    if (kernel_strsend(sock, msg) < 0) {
      free(msg);
      return -1;
    }
    free(msg);
    llprintf(TRACE_MSG, "OK\n");
    log = log_next(loglist);
  }

  return 1;
}


/* 
 * komunikace s monitorem
 */
int mn_recv(int sock) {
  char *msg;
  int op, i, n;
  char *tokens[10];
  int order;
  list_t list;

//  lprintf("mn, DEBUG: cekani na zpravu\n");
  if ((msg = kernel_strrecv(sock, TIMEOUT, 1)) == NULL) return -1;
  if (!msg[0]) return 0; /* timeout */
  llprintf(TRACE_MSG, "mn: prijata zprava '%s'\n", msg);

  op = 0;
  i = 0;
  while (msg[i] && msg[i] >= '0' && msg[i] <= '9') {
    op = op * 10 + (msg[i] - '0');
    i++;
  }

  switch (op) {
    case MSG_MN_USERS: {
      if ((n = msg_parse(msg, tokens, 10)) < 2) {
	printf("mn: chyba formatu zpravy MSG_MN_USERS(%s|%s...)\n", n > 0 ? tokens[0] : "(nic)", n > 1 ? tokens[1] : "(nic)");
	goto exc_free;
      }
      lprintf("mn: op = MSG_MN_USERS (0x%02X, order = '%s')\n", op, tokens[1]);

      /* druh trideni */
      order = parse_order(tokens[1]);

      free(msg);

      /* generovani setrideneho seznamu */
      if (!kernel_lock_read()) return -1;
      list_init(&list);
      for (i = 0; i < nusers; i++) {
        msg = msg_create(MSG_MN_USERS, "wwSSSS", uzivatele[i].id_user, uzivatele[i].id_group, uzivatele[i].jmeno, uzivatele[i].prijmeni, uzivatele[i].os_cislo, id2str(uzivatele[i].id_jis, sizeof(jis_t)));
        if (!msg) {
          printf("mn, vnitrni chyba: nelze vytvorit zpravu MSG_MN_USERS\n");
          list_done(&list);
          goto exc_unl;
        }
        switch (order) {
          case 1: list_insert_by_str(&list, msg, uzivatele[i].jmeno); break;
          case 2: list_insert_by_str(&list, msg, uzivatele[i].prijmeni); break;
          case 3: list_insert_by_str(&list, msg, uzivatele[i].os_cislo); break;
          case 4: list_insert_by_jis(&list, msg, uzivatele[i].id_jis); break;
          case 5: list_insert_by_int(&list, msg, uzivatele[i].id_user); break;
          case 6: list_insert_by_int(&list, msg, uzivatele[i].id_group); break;
	  default: list_add(&list, msg, NULL); break;
        }
      }
      if (!kernel_unlock_read()) {
        list_done(&list);
        return -1;
      }

      /* odeslani vsech pripravenych zprav */
      if (!list_send(sock, &list)) {
        list_done(&list);
        return -1;
      }
      list_done(&list);
      return 1;
    }

    case MSG_MN_LOGINS: {
      if ((n = msg_parse(msg, tokens, 10)) < 2) {
	printf("mn: chyba formatu zpravy MSG_MN_LOGINS(%s|%s...)\n", n > 0 ? tokens[0] : "(nic)", n > 1 ? tokens[1] : "(nic)");
	goto exc_free;
      }
      lprintf("mn: op = MSG_MN_LOGINS (0x%02X, order = '%s')\n", op, tokens[1]);

      /* druh trideni */
      order = parse_order(tokens[1]);

      free(msg);

      /* generovani setrideneho seznamu */
      if (!kernel_lock_read()) return -1;
      list_init(&list);
      for (i = 0; i < nusers; i++) {
        if (uzivatele[i].prihlasen) {
          msg = msg_create(MSG_MN_LOGINS, "wwSSSS", uzivatele[i].id_user, uzivatele[i].id_group, uzivatele[i].jmeno, uzivatele[i].prijmeni, uzivatele[i].os_cislo, id2str(uzivatele[i].id_jis, sizeof(jis_t)));
          if (!msg) {
            printf("mn, vnitrni chyba: nelze vytvorit zpravu MSG_MN_LOGINS\n");
            list_done(&list);
            goto exc_unl;
          }
          switch (order) {
            case 1: list_insert_by_str(&list, msg, uzivatele[i].jmeno); break;
            case 2: list_insert_by_str(&list, msg, uzivatele[i].prijmeni); break;
            case 3: list_insert_by_str(&list, msg, uzivatele[i].os_cislo); break;
            case 4: list_insert_by_jis(&list, msg, uzivatele[i].id_jis); break;
            case 5: list_insert_by_int(&list, msg, uzivatele[i].id_user); break;
            case 6: list_insert_by_int(&list, msg, uzivatele[i].id_group); break;
            default: list_add(&list, msg, NULL); break;
          }
	}
      }
      if (!kernel_unlock_read()) {
        list_done(&list);
        return -1;
      }

      /* odeslani vsech pripravenych zprav */
      if (!list_send(sock, &list)) {
        list_done(&list);
        return -1;
      }
      list_done(&list);
      return 1;
    }

    case MSG_MN_SENSORS: {
      lprintf("mn: op = MSG_MN_SENSORS (0x%02X)\n", op);
      if (!kernel_lock_read()) goto exc_free_fail;
      for (i = 0; i < nsensors; i++) {
	int nlogins, j;
	
        free(msg);
	nlogins = 0;
	for (j = 0; j < snimace[i].nusers; j++) {
	  if (snimace[i].users[j].info->prihlasen) nlogins++;
	}
        msg = msg_create(MSG_MN_SENSORS, "Hiwwwi", snimace[i].id, snimace[i].room, snimace[i].nusers, nlogins, snimace[i].nzones, snimace[i].active);
        if (!msg) {
          printf("mn, vnitrni chyba: nelze vytvorit zpravu MSG_MN_SENSORS\n");
	  goto exc_unl;
        }
	llprintf(TRACE_MSG, "mn: odesle se MSG_MN_SENSORS (%s)\n", msg);
        if (!kernel_strsend(sock, msg)) goto exc_free_fail;
      }
      if (!kernel_unlock_read()) goto exc_free_fail;
      free(msg);
      msg = msg_create(MSG_MN_END, "");
      if (!kernel_strsend(sock, msg)) goto exc_free_fail;
      break;
    }

    case MSG_MN_ROOM_LOGINS: {
      int room, cidlo, i;
      
      lprintf("mn: op = MSG_MN_ROOM_LOGINS (0x%02X)\n", op);
      if (sscanf(msg, "%d|%d", &op, &room) != 2) {
	printf("mn: chyba formatu zpravy MSG_MN_ROOM_LOGINS (%s)\n", msg);
	goto exc_free;
      }
      free(msg);
      /* hledani cidel v mistnosti */
      if (!kernel_lock_read()) return -1;
      for (cidlo = 0; cidlo < nsensors; cidlo++) {
        if (snimace[cidlo].room == room) {
          for (i = 0; i < snimace[cidlo].nusers; i++) {
	    user_info_t *user = snimace[cidlo].users->info + i;
	    
            if (user->prihlasen) {
              msg = msg_create(MSG_MN_ROOM_LOGINS, "wwSSSS", user->id_user, user->id_group, user->jmeno, user->prijmeni, user->os_cislo, id2str(user->id_jis, sizeof(jis_t)));
              if (!msg) {
                printf("mn, vnitrni chyba: nelze vytvorit zpravu MSG_MN_LOGINS\n");
                goto exc_unl;
              }
              if (!kernel_strsend(sock, msg)) goto exc_unl;
            }
          }
        }
      }
      if (!kernel_unlock_read()) return -1;
      /* koncova znacka */
      msg = msg_create(MSG_MN_END, "");
      if (!kernel_strsend(sock, msg)) goto exc_free_fail;
      break;
    }

    case MSG_MN_LOG: {
      int retval;
      
      lprintf("mn: op = MSG_MN_LOG (0x%02X)\n", op);
      if (!kernel_lock()) goto exc_free_fail;
      if ((retval = monitor_send_loglist(sock, &logs)) != 1) {
        if (!kernel_unlock()) return -1;
	return retval;
      }
      if ((retval = monitor_send_loglist(sock, &pending_logs)) != 1) {
        if (!kernel_unlock()) return -1;
	return retval;
      }
      free(msg);
      msg = msg_create(MSG_MN_END, "");
      llprintf(TRACE_MSG, "mn: odeslani konce (%s)\n", msg);
      if (kernel_strsend(sock, msg) < 0) goto exc_unl_fail;
      if (!kernel_unlock()) goto exc_free_fail;
      break;
    }

    case MSG_MN_SERVER_STATUS: {
      int i, active;

      lprintf("mn: op = MSG_MN_SERVER_STATUS (0x%02X)\n", op);
      free(msg);
      if (!kernel_lock_read()) goto exc_free_fail;
      active = 0;
      i = 0;
      while (i < nusers) {
        if (uzivatele[i].prihlasen) active++;
        i++;
      }
      msg = msg_create(MSG_MN_SERVER_STATUS, "wwSw", nusers, nsensors, kernel_get_uptime(), active);
      if (!kernel_unlock_read()) goto exc_free_fail;
      if (!msg) {
	printf("mn, vnitrni chyba: nelze vytvorit zpravu\n");
	return 0;
      }
      llprintf(TRACE_MSG, "mn: odeslani stavu (%s)\n", msg);
      if (!kernel_strsend(sock, msg)) goto exc_free_fail;
      break;
    }

    default:
      printf("mn: neznamy typ 0x%02X\n", op);
  }
  if (msg) free(msg);
  return 1;

exc_unl:
  if (!kernel_unlock_read()) goto exc_free_fail;
exc_free:
  free(msg);
  return 0;
exc_unl_fail:
  kernel_unlock_read();
exc_free_fail:
  free(msg);
  return -1;
}
#else


int main() {
  char *msg;
  char *tokens[100];
  int i, n;

  msg = msg_create(34, "wSS", 12345, "ahoj1", "ahoj2");
  printf("msg = %s\n", msg);
  n = msg_parse(msg, tokens, 100);
  for (i = 0; i < n; i++) printf("%02d: '%s'\n", i, tokens[i]);
  free(msg);

  return 1;
}
#endif
