/* sens.c */

/* Komunikace s modulem cidla. */

/* Poznamka:
 *   pri dotazu na konkretniho uzivatele i casovou zonu se pouziva vzdy agresivnejsi
 *   zamykani pro zapis i v pripade existence polozky v pameti (bez zmeny dat)
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "kernel.h"


/* 
 * cte hlavicku s poctem polozek,
 * jestlize je tam prave jedna polozka, vraci TRUE 
 *
 * (predpoklada jadro zamcene pro cteni)
 */
static int ziskej_pocet(int sock) {
  message_t *msg;
  int n;
  
  /*prijem poctu uzivatelu (0 = nenalezen, 1 = nalezen) */
  if((msg = kernel_recv(sock, TIMEOUT, 1)) == NULL) return 0;
  if (msg->op != MSG_DB_HEADER || !message_parse(msg, "W", &n)) {
    message_free(msg);
    return 0;
  }
  message_free(msg);

  /* kontrola konzistence */
  if (n != 0 && n != 1) {
    printf("sn, varovani: duplicitni zaznam v DB!\n");
    return 0;
  }

  return n == 1;
}


/* pomocna funkce pro hledani cidla z danou adresou */
static int hledej_cidlo(uint16_t adresa_sn) {
  int i, pozice = -1;
  
  for(i = 0; i < nsensors; i++) {
    if(snimace[i].id == adresa_sn) {
      pozice = i;
      break;
    }
  }

  return pozice;
}


static int pridej_cidlo(uint16_t adresa, int room, int active) {
  if (nsensors >= MAX_SENSORS) {
    printf("sn: uz je tu moc snimacu (maximum je %d)\n", MAX_SENSORS);
    return 0;
  }
  snimace[nsensors].id = adresa;
  snimace[nsensors].users = NULL;
  snimace[nsensors].tz = NULL;
  snimace[nsensors].nusers = 0;
  snimace[nsensors].nzones = 0;
  snimace[nsensors].room = room;
  snimace[nsensors].active = active;
  nsensors++;

  return 1;
}


/* nebyl-li nalezen uzivatel v tabulkach v jadru zepta se databaze
 * ta bud odpovi ze ma a tabulka useru je rozsirena, nebo odpovi ze nema -> fail
 * aresa snimace, id_user uzivatel, *radka pozice noveho uzivatele
 *
 * (predpoklada jadro zamcene pro zapis!)
 *
 */
static int zeptej_se_db_na_user(uint16_t adresa_sn, uint8_t id_user[], int *radka) {
  message_t *msg;
  int pozice=0;
  int id_zone, rights;
  jis_t id;
  uint8_t key[16];
  void *tmp;

  /* test databaze */
  if (sock_db == -1) {
    printf("sn: nelze se ptat na uzivatele, databaze neni pripojena\n");
    return 0;
  }

  /* odeslani dotazu na daneho uzivatele */
  msg = message_create(MSG_DB_REQ_ONE_USER, "WI", adresa_sn, id_user);
#ifdef TRACE_MSG
  message_dump("sn: MSG_DB_REQ_ONE_USER, ", msg);
#endif
  if(msg == NULL) {
    printf("sn: zpravu MSG_DB_REQ_ONE_USER se nepodarilo vytvorit.\n");
    return(0);
  }
  if (kernel_send(sock_db, msg) < 0) {
    printf("sn: nepodarilo se odeslat zpravu MSG_DB_REQ_ONE_USER.\n");
    message_free(msg);
    return(0);
  }
  message_free(msg);

  /* zjisteni, zda uzivatel existuje */
  if (!ziskej_pocet(sock_db)) {
    lprintf("Uzivatel u snimace %04X nenalezen\n", adresa_sn);
    return 0;
  }

  /* prijem noveho uzivatele (s kontrolou) */
  if((msg = kernel_recv(sock_db, TIMEOUT, 1)) == NULL) return 0;
  if (msg->op == MSG_NONE) {
    message_free(msg);
    return 0;
  }
  if(msg->op != MSG_DB_USER || !message_parse(msg, "IBW", id, key, &id_zone, &rights)) {
    printf("sn: prijata chybna zprava MSG_DB_USER (typ %02X, prijato %02X).\n", MSG_DB_USER, msg->op);
    message_free(msg);
    return(0);
  }
#if TRACE_MSG
  message_dump("sn: prijato ", msg);
#endif

  /*nalezeni snimace*/
  if ((pozice = hledej_cidlo(adresa_sn)) == -1) {
    kernel_oops("sn, vnitrni chyba: snimac %04X nenalezen!\n", adresa_sn);
    return 0;
  }

  /*realokace tabulky user*/
  tmp = realloc(snimace[pozice].users, (snimace[pozice].nusers+1) * sizeof(user_t));
  if (!tmp) {
    kernel_oops("sn: neuspesny realloc() uzivatelu!\n");
    return 0;
  }
  snimace[pozice].users = tmp;
  snimace[pozice].nusers++;
  
  /*zkopirovani novych udaju do user*/
  snimace[pozice].users[snimace[pozice].nusers-1].id_zone = id_zone;
  snimace[pozice].users[snimace[pozice].nusers-1].rights = rights;

  if ((snimace[pozice].users[snimace[pozice].nusers-1].info = kernel_search_user(id)) == NULL) {
    printf("db: kontrola pridani uplne nove uzivatele do databaze jeste neni implementovana!\n");
    snimace[pozice].nusers--;
    return 0;
  }

  /*radka = cislo radky, kde se nachazi nove pridany uzivatel*/
  *radka = snimace[pozice].nusers-1;

  return 1;
}


/*
 *
 * (predpoklada jadro zamcene pro zapis!) 
 */
static int zeptej_se_db_na_tz(uint16_t adresa_sn, uint8_t id_zone, int *radka) {
  message_t *msg;
  int pozice;
  uint8_t bmp[BITMAP_SIZE];
  void *tmp;
  int foo;

  /* test databaze */
  if (sock_db == -1) {
    printf("sn: nelze se ptat na casovou zonu, databaze neni pripojena\n");
    return 0;
  }

  /* odeslani dotazu na danou casovou zonu */
  msg = message_create(MSG_DB_REQ_ONE_TIME_ZONE, "WB", adresa_sn, id_zone);
#ifdef TRACE_MSG
  message_dump("sn: MSG_DB_REQ_ONE_TIME_ZONE, ", msg);
#endif
  if(msg == NULL) {
    printf("sn: zpravu MSG_DB_REQ_ONE_TIME_ZONE se nepodarilo vytvorit.\n");
    return(0);
  }
  if (kernel_send(sock_db, msg) < 0) {
    printf("sn: nepodarilo se odeslat zpravu MSG_DB_REQ_ONE_TIME_ZONE.\n");
    message_free(msg);
    return(0);
  }
  message_free(msg);

  /* zjisteni, zda casova zona existuje */
  if (!ziskej_pocet(sock_db)) {
    lprintf("Casova zona %d u snimace %04X nenalezen\n", id_zone, adresa_sn);
    return 0;
  }

  /* prijem nove zony (s kontrolou) */
  if((msg = kernel_recv(sock_db, TIMEOUT, 1)) == NULL) return 0;
  if (msg->op == MSG_NONE) {
    message_free(msg);
    return 0;
  }
  if(msg->op != MSG_DB_TIME_ZONE || !message_parse(msg, "BT", &foo, bmp)) {
    printf("sn: prijata chybna zprava MSG_DB_TIME_ZONE (typ %02X, prijato %02X).\n", MSG_DB_TIME_ZONE, msg->op);
    message_free(msg);
    return(0);
  }
  message_free(msg);
  if (foo != id_zone) {
    printf("sn: z databaze vracena chybna data, ocekavana zona %" PRIu8 ", vracena %d\n", id_zone, foo);
    return 0;
  }

  /*nalezeni snimace*/
  if ((pozice = hledej_cidlo(adresa_sn)) == -1) {
    kernel_oops("sn, vnitrni chyba: snimac %04X nenalezen!\n", adresa_sn);
    return 0;
  }
  
  /*realokace tabulky casovych zon*/
  tmp = realloc(snimace[pozice].tz, (snimace[pozice].nzones+1) * sizeof(time_zone_t));
  if (!tmp) {
    kernel_oops("sn: neuspesny realloc() casovych zon!\n");
    return 0;
  }
  snimace[pozice].tz = tmp;
  snimace[pozice].nzones++;

  /*zkopirovani novych udaju do snimace->tz*/
  snimace[pozice].tz[snimace[pozice].nzones-1].id = id_zone;
  memcpy(snimace[pozice].tz[snimace[pozice].nzones-1].bmp, bmp, sizeof(bmp));

  /*radka = cislo radky, kde se nachazi nove pridana casova zona*/
  *radka = snimace[pozice].nzones-1;

  return 1;
}


/*porovnava dve pole cisel, pri schode vraci 1, jinak 0*/
static int equals(uint8_t tab_1[], uint8_t tab_2[]) {
  int i;
  for(i = 0; i < 8; i++) {
    if(tab_1[i] != tab_2[i])
      return(0);
  }
  return(1);
}


int sn_recv(int sock_sn) {
  message_t *msg;
  static int state = 0;
  int adresa_sn;
  uint8_t id_zony;
  jis_t id_user;
  int retval, foo2, i;
  int pozice;
  int radka=0;

  /*prijem zpravy od modulu cidel*/
  if((msg = kernel_recv(sock_sn, TIMEOUT, 1)) == NULL) return -1;
#if TRACE_MSG
  message_dump("sn: prijata zprava, ", msg);
#endif

  switch(msg->op) {
/*******************************************************************************************/    
    /*zadost o tabulku TIME ZONES*/
    case MSG_SN_REQUEST_TIME_ZONES: {
      lprintf("sn: MSG_SN_REQUEST_TIME_ZONES\n");
      if (state < 2) /*ok*/state++;
      
      /*zjisteni adresy snimace*/
      retval = message_parse(msg, "W", &adresa_sn);
      message_free(msg);
      if(retval == 0) {
        printf("sn: prijata zprava, spatny format, neodpovida MSG_SN_REQUEST_TIME_ZONE.\n");
        /*
         *fail
         */
        return(0);
      }

      /* cteni dat */
      if (!kernel_lock_read()) return -1;
      
      /*nalezeni snimace v tabulce snimacu*/
      if((pozice = hledej_cidlo(adresa_sn)) == -1) {/*snimac nenalezen*/
        printf("sn: snimac %04X v tabulce nebyl nalezen, spatna adresa snimace.\n", adresa_sn);
        /*
         *fail
         */
	kernel_unlock_read();
	return 0;
      }
      /*sestaveni header*/
      msg = message_create(MSG_SN_HEADER, "W", snimace[pozice].nzones);
#ifdef TRACE_MSG
      message_dump("sn: posle se MSG_SN_HEADER(TZ), ", msg);
#endif
      if (msg == NULL) {
        printf("sn: Chyba pri sestavovani MSG_SN_HEADER.\n");
        /*
         *fail
         */
	kernel_unlock_read();
        return(0);
      }
      /*odeslani header*/
      if (kernel_send(sock_sn, msg) < 0) {
        printf("sn: chyba pri odesilani ramce MSG_SN_HEADER.\n");
        /*
         *fail
         */
        message_free(msg);
        if (!kernel_unlock_read()) return -1;
        return(0);
      }
      llprintf(TRACE_MSG, "OK\n");
      
      /*zruseni odeslane zpravy*/
      message_free(msg);
      
      /*odeslani time zones*/
      for(i = 0; i < snimace[pozice].nzones; i++) {
        msg = message_create(MSG_SN_TIME_ZONE, "BT", snimace[pozice].tz[i].id, snimace[pozice].tz[i].bmp);
        if (msg == NULL) {
          printf("sn: Chyba pri sestavovani MSG_SN_TIME_ZONE.\n");     
          /*
           *fail
           */
	  kernel_unlock_read();
          return(0);
        }
#ifdef TRACE_MSG
        message_dump("sn: MSG_SN_TIME_ZONE, ", msg);
#endif
        
        if (kernel_send(sock_sn, msg) < 0) {
          printf("sn: chyba pri odesilani ramce MSG_SN_TIME_ZONE.\n");
          /*
           *fail
           */
          message_free(msg);
	  kernel_unlock_read();
          return(0);
        }
        lprintf("sn: zprava MSG_SN_TIME_ZONE odeslana.\n");
        message_free(msg);
      }/*for*/

      /* skoncilo cteni dat */
      if (!kernel_unlock_read()) return -1;

      break;/*konec*/
    }
    
/*******************************************************************************************/    
    /*zadost o tabulku useru*/
    case MSG_SN_REQUEST_USERS: {
      lprintf("sn: MSG_SN_REQUEST_TIME_USERS\n");
      if (state < 2) /*ok*/state++;
      /*zjisteni adresy snimace*/
      retval = message_parse(msg, "W", &adresa_sn);
      message_free(msg);
      if(retval == 0) {
        printf("sn: prijata zprava, spatny format, neodpovida MSG_SN_REQUEST_USERS.\n");
        /*
         *fail
         */
        return(0);
      }

      /* cteni dat  */
      if (!kernel_lock_read()) return -1;

      /*nalezeni snimace v tabulce snimacu*/
      if((pozice = hledej_cidlo(adresa_sn)) == -1) {/*snimac nenalezen*/
        printf("sn: snimac %04X nebyl nalezen, spatna adresa snimace.\n", adresa_sn);
        /*
         *fail
         */
	kernel_unlock_read();
        return(0);
      }
      /*sestaveni header*/
      msg = message_create(MSG_SN_HEADER, "W", snimace[pozice].nusers);
      if (msg == NULL) {
        printf("sn: Chyba pri sestavovani MSG_SN_HEADER.\n");
        /*
         *fail
         */
	kernel_unlock_read();
        return(0);
      }
#ifdef TRACE_MSG
      message_dump("sn: MSG_SN_HEADER(users), ", msg);
#endif
      /*odeslani header*/
      if (kernel_send(sock_sn, msg) < 0) {
        printf("sn: chyba pri odesilani ramce MSG_SN_HEADER.\n");
        /*
         *fail
         */
        message_free(msg);
	kernel_unlock_read();
        return(0);
      }
      
      /*zruseni prijate zpravy*/
      message_free(msg);
      
      /*odeslani users*/
      for(i = 0; i < snimace[pozice].nusers; i++) {
        msg = message_create(MSG_SN_USER, "IKBW", snimace[pozice].users[i].info->id_jis, snimace[pozice].users[i].info->key, snimace[pozice].users[i].id_zone, snimace[pozice].users[i].rights);
        if (msg == NULL) {
          printf("sn: Chyba pri sestavovani MSG_SN_USER.\n");     
          /*
           *fail
           */
	  kernel_unlock_read();
          return(0);
        }
#ifdef TRACE_MSG
        message_dump("sn: MSG_SN_USERS, ", msg);
#endif
        
        if (kernel_send(sock_sn, msg) < 0) {
          printf("sn: chyba pri odesilani ramce MSG_SN_USER.\n");
          /*
           *fail
           */
          message_free(msg);
	  kernel_unlock_read();
          return(0);
        }
        #ifdef LOG
          printf("sn: zprava MSG_SN_USER odeslana.\n");
        #endif
        message_free(msg);

      }/*for*/

      /* skoncilo cteni dat */
      if (!kernel_unlock_read()) return -1;

      break;/*konec*/
    }

/*******************************************************************************************/    
    /*zadost o jednu casovou zonu pro snimac popsany adresou a zonu pospanou id_zony*/
    case MSG_SN_REQUEST_1_TIME_ZONE: {
      lprintf("sn: MSG_SN_REQUEST_1_TIME_ZONE\n");
      if(state < 2) /*ko*/ return 0;
      /*zjisteni adresy snimace a id zony z prijate zpravy*/
      retval = message_parse(msg, "WB", &adresa_sn, &foo2);
      message_free(msg);
      if(retval == 0) {
        printf("sn: prijata zprava, spatny format, neodpovida MSG_SN_1_TIME_ZONE.\n");
        /*
         *fail
         */
        return(0);
      }
      id_zony = foo2;

      /* cteni i zapis dat */
      if (!kernel_lock()) return -1;
     
      /*hledani snimace v tabulce snimacu*/
      if((pozice = hledej_cidlo(adresa_sn)) == -1) {/*snimac nenalezen*/
        printf("sn: chyba, adresa snimace nenalezena (MSG_SN_REQUEST_1_TIME_ZONE).\n");
        kernel_unlock();
        return(0);
      }
      /*snimac byl nalezen*/
      /*hledani casove zony v tabulce casovych zon pro nalezeny snimac*/
      radka = -1;
      for(i = 0; i < snimace[pozice].nzones; i++) {
        if(snimace[pozice].tz[i].id == id_zony) {/*id zony nalezeno*/
          radka = i;
          break;
        }
      }
      if(radka == -1) { /*id zony nenalezeno*/
        lprintf("sn: zona v tabulce nenalezena (MSG_SN_REQUEST_1_TIME_ZONE), dotaz na DB\n");

        /*zkusim se zeptat databaze*/
        if (!zeptej_se_db_na_tz(adresa_sn, id_zony, &radka)) {
        /*TZ nebyla nalezena ani v databazi, poslu MSG_SN_NOT_FOUND*/
          msg = message_create(MSG_SN_NOT_FOUND, "");
          if(msg == NULL) {
            printf("sn: chyba, zpravu MSG_SN_NOT_FOUND se nepodarilo vytvorit.\n");
            kernel_unlock();
            return 0;
          } 
#ifdef TRACE_MSG
          message_dump("sn: MSG_SN_NOT_FOUND, ", msg);
#endif
          if(kernel_send(sock_sn, msg) < 0) {
            printf("sn: zpravu MSG_SN_NOT_FOUND se nepodarilo odeslat.\n");
            message_free(msg);
            kernel_unlock();
            return 0;
          }
          lprintf("sn: zprava MSG_SN_NOT_FOUND odeslana.\n");
          message_free(msg);
          kernel_unlock();
          return 0;
        }
      }
      /*jestlize retval = 1, uzivatel byl v databazi nalezen a cislo kde se nechazi 
       * v snimace[pozice]->user[...] je uvedeno v promenne 'radka'*/
      /*vytvoreni zpravy s nalezenou casovou zonou*/
      msg = message_create(MSG_SN_TIME_ZONE, "BT", snimace[pozice].tz[radka].id, snimace[pozice].tz[radka].bmp);
      if(msg == NULL) {/*zpravu se nepodarli vytvorit*/
        printf("sn: chyba, zpravu MSG_SN_TIME_ZONE se nepodarilo vytvorit, (odpoved na MSG_SN_REQUEST_1_TIME_ZONE).\n");
        kernel_unlock();
        return(0);
      }
#ifdef TRACE_MSG
      message_dump("sn: MSG_SN_TIME_ZONE, ", msg);
#endif
      /*odeslani zpravy*/
      if (kernel_send(sock_sn, msg) < 0) {
        printf("sn: chyba pri odesilani ramce MSG_SN_TIME_ZONE, (odpoved na MSG_SN_REQUEST_1_TIME_ZONE).\n");
        /*
         *fail
         */
        message_free(msg);
        kernel_unlock();
        return(0);
      }
      lprintf("sn: zprava MSG_SN_TIME_ZONE odeslana.\n");
      message_free(msg);
      if (!kernel_unlock()) return -1;
      break;
    }

/*******************************************************************************************/   
/*zadost o jednoho uzivatele pro snimac popsany adresou a uzivatele popsanehp id uzivatele*/
    case MSG_SN_REQUEST_1_USER: {
     lprintf("sn: MSG_SN_REQUEST_1_USER\n");
     if(state < 2) /*ko*/ return 0;
     /*zjisteni adresy snimace a id usera z prijate zpravy*/
     if(!message_parse(msg, "WI", &adresa_sn, id_user)) {
        printf("sn: prijata zprava, spatny format, neodpovida MSG_SN_1_USER.\n");
        /*
         *fail
         */
        message_free(msg);
        return(0);
     }
     message_free(msg);
     
     /* cteni i zapis dat */
     if (!kernel_lock()) return -1;

     /*hledani snimace v tabulce snimacu*/
     
     pozice = -1;
     if ((pozice = hledej_cidlo(adresa_sn)) == -1) {/*snimac nenalezen*/
       printf("sn: chyba, adresa snimace nenalezena (MSG_SN_REQUEST_1_USER).\n");
       /*
        * fail
        */
       kernel_unlock();
       return(0);
     }
     /*snimac byl nalezen */
     /*hledani usera v tabulce useru pro nalezeny snimac*/
     radka = -1;
     for(i = 0; i < snimace[pozice].nusers; i++) {
       if(equals(snimace[pozice].users[i].info->id_jis, id_user)) { /*id usera nalezeno*/
         radka = i;
         break;
       }
     }
     if(radka == -1) { /*id zony nenalezeno*/
       lprintf("sn: user v tabulce nenalezen (MSG_SN_REQUEST_1_USER), dotaz na DB.\n");
       /*zkusim se zeptat databaze*/
       if(!zeptej_se_db_na_user(adresa_sn, id_user, &radka)) {
       /*uzivatel nebyl nalezen ani v databazi, poslu MSG_SN_NOT_FOUND*/
         msg = message_create(MSG_SN_NOT_FOUND, "");
         if(msg == NULL) {
           printf("sn: chyba, zpravu MSG_SN_NOT_FOUND se nepodarilo vytvorit.\n");
           kernel_unlock();
           return 0;
         }
#ifdef TRACE_MSG
         message_dump("sn: MSG_SN_NOT_FOUND, ", msg);
#endif
         if(kernel_send(sock_sn, msg) < 0) {
           printf("sn: zpravu MSG_SN_NOT_FOUND se nepodarilo odeslat.\n");
           message_free(0);
           kernel_unlock();
           return 0;
         }
         lprintf("sn: zprava MSG_SN_NOT_FOUND odeslana.\n");
         message_free(msg);
         kernel_unlock();
         return 0;
       }
     }
     /*jestlize retval = 1, uzivatel byl v databazi nalezen a cislo kde se nechazi 
      * v snimace[pozice]->user[...] je uvedeno v promenne 'radka'*/
     /*vytvoreni zpravy s nalezenym userem*/
     msg = message_create(MSG_SN_USER, "IKBW", snimace[pozice].users[radka].info->id_jis, snimace[pozice].users[radka].info->key, snimace[pozice].users[radka].id_zone, snimace[pozice].users[radka].rights);
     if(msg == NULL) {/*zpravu se nepodarli vytvorit*/
       printf("sn: chyba, zpravu MSG_SN_USER se nepodarilo vytvorit, (odpoved na MSG_SN_REQUEST_1_USER).\n");
       kernel_unlock();
       return(0);
     }
#ifdef TRACE_MSG
      message_dump("sn: MSG_SN_USER, ", msg);
#endif
     /*odeslani zpravy*/
     if (kernel_send(sock_sn, msg) < 0) {
       printf("sn: chyba pri odesilani zpravy MSG_SN_USER, (odpoved na MSG_SN_REQUEST_1_USER).\n");
       /*
        *fail
        */
       message_free(msg);
       kernel_unlock();
       return(0);
     }
     lprintf("sn: zprava MSG_SN_USER odeslana.\n");
     message_free(msg);
     if (!kernel_unlock()) return -1;
     break;
    }
/*******************************************************************************************/    
    case MSG_SN_LOG: {
      int cidlo, type;
      jis_t user;
      char cas[17];
      char *text;
      user_info_t *ui;

      if (!message_parse(msg, "WIKB", &cidlo, user, cas, &type)) {
	message_free(msg);
	return 0;
      }
      message_free(msg);
      lprintf("sn: MSG_SN_LOG(cidlo 0x%04X, cas %s, typ 0x%02X)\n", cidlo, cas, type);

      if (!kernel_lock()) return -1;
      /* najde cidlo */
      if ((pozice = hledej_cidlo(cidlo)) == -1) {
        printf("sn: cidlo s adresou %04X nenalezeno (MSG_SN_LOG)\n", cidlo);
        goto exc_log_fail;
      }
      /* najde info uzivatele */
      if ((ui = kernel_search_user(user)) == NULL) {
        printf("sn: uzivatel s JIS %s nenalezen\n", id2str(user, 8));
        log_add(&pending_logs, cidlo, 0, user, cas, "uzivatel neni v databazi");
        goto exc_log_fail;
      }
      /* odhlasi uzivatele (treba kdyz je jinde) */
      kernel_logout_user(user);
      switch (type) {
	case 0x62:
          text = "pristup povolen";
          if (!kernel_login_user(pozice, user)) {
            printf("Uzivatel %s nenalezen v cidle %04X!\n", ui->os_cislo, snimace[pozice].id);
            log_add(&pending_logs, cidlo, 0, user, cas, "uzivatel v cidle nenalezen");
            goto exc_log_fail;
          }
          break;
	default: text = "pristup zakazan";
      }
      log_add(&pending_logs, cidlo, type, user, cas, text);
      if (!kernel_unlock()) return -1;
      break;
exc_log_fail:
      if (!kernel_unlock()) return -1;
      return 0;
    }

  case MSG_SN_WC: {
    int cidlo, pos;
    
    if (!message_parse(msg, "W", &cidlo)) {
      message_free(msg);
      return 0;
    }
    message_free(msg);
    lprintf("sn: MSG_SN_WC(0x%04X)\n", cidlo);
    if (!kernel_lock()) return -1;
    if ((pos = hledej_cidlo(cidlo)) == -1) {
      printf("sn, varovani: neaktivni cidlo %04X neni v databazi, pridano aspon do jadra\n", cidlo);
      if (!pridej_cidlo(cidlo, -1, 0)) {
        if (!kernel_unlock()) return -1;
        return 0;
      }
    } else snimace[pos].active = 0;
    if (!kernel_unlock()) return -1;
    break;
  }

  case MSG_SN_OK: {
    int cidlo, pos;
    
    if (!message_parse(msg, "W", &cidlo)) {
      message_free(msg);
      return 0;
    }
    message_free(msg);
    lprintf("sn: MSG_SN_OK(0x%04X)\n", cidlo);
    if (!kernel_lock()) return -1;
    if ((pos = hledej_cidlo(cidlo)) == -1) {
      printf("sn, varovani: aktivni cidlo %04X neni v databazi, pridano aspon do jadra\n", cidlo);
      if (!pridej_cidlo(cidlo, -1, 1)) {
        if (!kernel_unlock()) return -1;
        return 0;
      }
    } else snimace[pos].active = 1;
    if (!kernel_unlock()) return -1;
    break;
  }

  case MSG_NONE: {
    lprintf("sn: timeout\n");
    message_free(msg);
    /* pri priznaku zneplatneni dat se posle MSG_SN_RENEW a priznak se schodi,
     * pri neuspechu se necha a sn_recv() vraci 0
     */
    if (sn_invalidate) {
      printf("sn: zneplatneni dat snimace\n");
      if (!kernel_lock()) return -1;
      if (kernel_send_simple_message(sock_sn, MSG_SN_RENEW)) sn_invalidate = 0;
      if (!kernel_unlock()) return -1;
      if (sn_invalidate) return 0;
    }
    return 0;
  }

   default:
     printf("sn: Prijata nezpracovavana zprava od snimace, typ %02X\n.", msg->op);
     message_free(msg);
      
  }/*switch*/

  return 1;
}
